# coding: utf8
from __future__ import unicode_literals


ADVERBS = set("""
'tween a.d. a.k.a. a.m. aback abaft abaxially abeam abed abjectly ably
abnormally aboard abominably aborad abortively about above aboveboard abreast
abroad abruptly absently absentmindedly absolutely abstemiously abstractedly
abstractly abstrusely absurdly abundantly abusively abysmally academically
accelerando acceptably accidentally accommodatingly accordingly accurately
accusingly acoustically across actively actually acutely ad adagio adamantly
adaxially additionally adequately adjectivally adjectively administratively
admirably admiringly admittedly adorably adoringly adrift adroitly adulterously
advantageously adverbially adversely advertently advisedly aerially
aesthetically afar affably affectedly affectingly affectionately affirmatively
afield afoot afresh aft after afterward afterwards again aggravatingly
aggressively agilely ago agonizingly agreeably aground ahead ahorse ahorseback
aimlessly airily akimbo alarmingly alas alee alertly alfresco algebraically
alias alike all all-fired allegedly allegorically allegretto allegro
alliteratively allowably almost aloft alone along alongside aloof aloud
alphabetically already alright also alternately alternatively altogether
altruistically always amain amateurishly amazingly ambiguously ambitiously
amiably amicably amidship amidships amiss amok amorally amorously amply amuck
amusingly anachronistically analogously analytically anarchically anatomically
anciently andante anew angelically angrily animatedly anisotropically annoyingly
annually anomalously anon anonymously antagonistically antecedently anteriorly
anticlockwise antithetically anxiously any anyhow anymore anyplace anyway
anyways anywhere apace apart apathetically apiece apologetically appallingly
apparently appealingly appositively appreciably appreciatively apprehensively
appropriately approvingly approximately apropos aptly arbitrarily
architecturally archly ardently arduously arguably argumentatively aright
aristocratically arithmetically around arrogantly artfully articulately
artificially artistically artlessly as asap ascetically asea asexually ashamedly
ashore aside askance askew aslant asleep assertively assiduously assuredly
astern astonishingly astraddle astray astride astronomically astutely asunder
asymmetrically asymptotically athwart atonally atop atrociously attentively
attractively attributively atypically audaciously audibly aurally auspiciously
austerely authentically authoritatively autocratically automatically
avariciously avidly avowedly away awful awfully awhile awkwardly awry axially
axiomatically b.c. b.c.e. baby-like baby-wise back backstage backward backwards
bacterially bad badly baldly balefully balmily banefully bang bannerlike
banteringly barbarously bareback barebacked barefacedly barefoot barefooted
barely basely bashfully basically bawdily bc bce beastly beautifully becomingly
befittingly before beforehand behind behindhand belatedly believably believingly
belike belligerently below beneath beneficially benevolently benignantly
benignly beseechingly besides best bestially betimes better between betwixt
bewilderedly bewilderingly bewitchingly beyond biannually biennially big
bilaterally bilingually bimonthly binaurally biochemically biologically
biradially bitingly bitter bitterly biweekly biyearly blamelessly blandly
blankly blasphemously blatantly bleakly blessedly blindly blissfully blithely
bloodily bloodlessly bloody bluffly bluntly boastfully bodily boiling
boisterously boldly bolt bombastically bonnily boorishly boringly boundlessly
bounteously bountifully boyishly boylike brashly bravely brazenly breadthways
breadthwise breast-deep breast-high breathlessly breezily briefly bright
brightly brilliantly briskly broad-mindedly broadly broadside broadwise
brotherly brusquely brutally brutishly bumptiously buoyantly bureaucratically
busily but buxomly by c.e. c.o.d. cagily cajolingly calculatingly callously
calmly calumniously candidly cannily canonically cantankerously cap-a-pie
capably capriciously captiously captivatingly carefully carelessly carnally
casually catalytically catastrophically categorically caudal caudally causally
caustically cautiously cavalierly ce ceaselessly centennially centrally
cerebrally ceremonially ceremoniously certainly cf cf. chaotically
characteristically charily charitably charmingly chastely chattily cheaply
cheekily cheerfully cheerily cheerlessly chemically chiefly childishly chintzily
chirpily chivalrously chock chock-a-block chop-chop chorally chromatically
chromatographically chronically chronologically churlishly circularly
circumspectly circumstantially civilly clammily clamorously clannishly
classically clean cleanly clear clearly cleverly climatically clinically
cliquishly clockwise close closely closer closest cloyingly clumsily coarsely
coastward coastwise coaxingly cod cognitively coherently coincidentally
coincidently cold-bloodedly coldly collect collectedly collectively colloidally
colloquially combatively comfortably comfortingly comically commendable
commensally commercially commonly communally compactly comparably comparatively
compassionately compatibly competently competitively complacently complainingly
completely complexly composedly comprehensively compulsively compulsorily
computationally con concavely conceitedly conceivably conceptually concernedly
concisely conclusively concretely concurrently condescendingly conditionally
confessedly confidentially confidently confidingly conformably confoundedly
confusedly confusingly congenially conically conjecturally conjointly conjugally
connubial conscientiously consciously consecutive consecutively consequentially
consequently conservatively considerably considerately consistently consolingly
conspicuously constantly constitutionally constrainedly constructively
contagiously contemporaneously contemptibly contemptuously contentedly
contextually continually continuously contractually contradictorily contrarily
contrariwise contrastingly contritely controversially contumaciously
contumeliously conveniently conventionally conversationally conversely convexly
convincingly convivially convulsively coolly cooperatively coordinately
copiously coquettishly cordially correctly correspondingly corruptedly corruptly
cortically cosily cosmetically coterminously counter counteractively
counterclockwise counterintuitively courageously course courteously covertly
covetously coyly cozily craftily crazily creakily creakingly creatively credibly
creditably credulously criminally crisply crisscross critically crookedly cross-
country cross-legged cross-linguistically crossly crosstown crossways crosswise
crucially crudely cruelly crushingly cryptically cryptographically culpably
culturally cumulatively cunningly curiously currently currishly cursedly
cursively cursorily curtly curvaceously cussedly customarily cutely cuttingly
cynically cytophotometrically cytoplasmically daftly daily daintily damn
damnably damned damply dandily dangerously daringly darkly dashingly dauntingly
dauntlessly daylong dazedly dazzlingly dead deadly deadpan dear dearly deathly
deceitfully deceivingly decent decently deceptively decidedly decipherably
decisively decoratively decorously deep deeply defectively defenceless
defencelessly defenseless defenselessly defensively deferentially defiantly
definitely deftly dejectedly deliberately delicately deliciously delightedly
delightfully deliriously delusively demandingly demeaningly dementedly
democratically demoniacally demonstrably demonstratively demurely
denominationally densely departmentally dependably deplorably deprecatively
depressingly derisively derisorily descriptively deservedly designedly
desolately despairingly desperately despicably despitefully despondently
destructively determinedly detestably detrimentally deucedly developmentally
devilish devilishly deviously devotedly devoutly dexterously dextrously
diabolically diagonally diagrammatically dialectically diametrically
dichotomously dictatorially didactically differentially differently diffidently
diffusely digitally digitately diligently dimly dingdong dingily diplomatically
direct directly direfully dirtily disadvantageously disagreeably disappointedly
disappointingly disapprovingly disastrously disbelievingly disconcertingly
disconsolately discontentedly discordantly discouragingly discourteously
discreditably discreetly discursively disdainfully disgracefully disgustedly
disgustingly dishonestly dishonorably dishonourably disingenuously
disinterestedly disjointedly disloyally dismally disobediently disparagingly
dispassionately dispiritedly displeasingly disproportionately disputatiously
disquietingly disregarding disregardless disreputably disrespectfully
disruptively dissolutely distally distantly distastefully distinctively
distinctly distractedly distressfully distressingly distributively distrustfully
disturbingly diversely divertingly divinely dizzily doctrinally doggedly doggo
dogmatically dolce dolefully doltishly domestically domineeringly dorsally
dorsoventrally dottily double doubly doubtfully doubtless doubtlessly dourly
dowdily down downfield downhill downright downriver downstage downstairs
downstream downtown downward downwardly downwards downwind drably draggingly
dramatically drastically dreadfully dreamfully dreamily drearily drily drippily
dripping droopingly drop-dead drowsily drunkenly dryly dubiously due dully duly
dumbly dutifully dynamically e'en e'er e.g. each eagerly earlier earliest early
earnestly easily east easterly eastward eastwards easy ebulliently eccentrically
ecclesiastically ecologically economically ecstatically edgeways edgewise
editorially educationally eerily effectively effectually efficaciously
efficiently effortlessly effusively egotistically either elaborately
electrically electronically electrostatically elegantly elementarily eloquently
elsewhere embarrassingly eminently emotionally empathetically emphatically
empirically emulously enchantingly encouragingly end-to-end endearingly
endlessly endogenously enduringly endways endwise energetically engagingly
enigmatically enjoyably enormously enough enquiringly enterprisingly
entertainingly enthrallingly enthusiastically entirely entreatingly enviably
enviously environmentally episodically equably equally equitably equivocally
erectly ergo erotically erratically erroneously erst erstwhile eruditely
eschatologically especially essentially esthetically etc. etcetera eternally
ethically ethnically euphemistically evasively even evenhandedly evenly
eventually ever everlastingly evermore everyplace everywhere evidently evilly
evolutionarily exactly exaggeratedly exasperatingly exceedingly excellently
exceptionally excessively excitedly excitingly exclusively excruciatingly
excusably exhaustively exorbitantly expansively expectantly expediently
expeditiously expensively experimentally expertly explicitly explosively
exponentially express expressively expressly exquisitely extemporaneously
extemporarily extempore extensively externally extortionately extra
extraordinarily extravagantly extremely exuberantly exultantly exultingly
fabulously face-to-face facetiously facially factually faddily faddishly fain
faintly fair fairly faithfully faithlessly false falsely falteringly familiarly
famously fanatically fancifully fantastically far farcically farther farthest
fascinatingly fashionably fast faster fastest fastidiously fatally fatefully
fatuously faultily faultlessly favorably favourably fearfully fearlessly
fearsomely feasibly fecklessly federally feebly feelingly feetfirst felicitously
ferociously fervently fervidly feudally feverishly fictitiously fiendishly
fiercely fierily fifthly figuratively filthily finally financially fine finely
finitely firm firmly first first-rate firsthand firstly fiscally fishily
fitfully fitly fittingly fixedly flabbily flagrantly flamboyantly flashily flat
flatly flawlessly fleetly flexibly flimsily flip-flap flippantly flirtatiously
flop floridly fluently flush focally fondly foolishly forbiddingly forcefully
forcibly fore foremost forever forevermore forgetfully forgivably forgivingly
forlornly formally formerly formidably formlessly forrad forrader forrard
forsooth forte forth forthright forthrightly forthwith fortissimo fortnightly
fortuitously fortunately forward forwards foully fourfold foursquare fourth
fourthly foxily fractiously frankly frantically fraternally fraudulently
freakishly free freely frenetically frenziedly frequently fresh freshly
fretfully frighteningly frightfully frigidly friskily frivolously frontally
frontward frontwards frostily frothily frowningly frugally fruitfully
fruitlessly frumpily frumpishly fucking fugally full full-time fully fulsomely
functionally fundamentally funnily furiously further furthermore furthest
furtively fussily futilely gaily gainfully gainlessly gallantly gamely garishly
garrulously gaudily gayly genealogically generally generically generously
genetically genially genteelly gently genuinely geographically geologically
geometrically geothermally gibingly giddily gingerly girlishly glacially gladly
glaringly gleefully glibly glissando gloatingly globally gloomily gloriously
glossily gloweringly glowingly glumly gluttonously goddam goddamn goddamned good
good-naturedly gorgeously governmentally gracefully gracelessly graciously
gradually grammatically grandiloquently grandiosely grandly graphically
gratefully gratifyingly gratingly gratis gratuitously gravely gravitationally
grayly greasily greatly greedily greenly gregariously greyly grievously grimly
gropingly grossly grotesquely grouchily grubbily grudgingly gruesomely gruffly
grumpily grungily guardedly guiltily gushingly gutturally habitually haggardly
half half-and-half half-heartedly half-hourly half-price half-time half-yearly
halfway haltingly handily handsomely haphazard haphazardly haply happily
haptically hard hardly harmfully harmlessly harmonically harmoniously harshly
harum-scarum hastily hatefully haughtily hazardously hazily head-on head-to-head
headfirst headlong healthily heaps heartily heartlessly heatedly heavenward
heavenwardly heavenwards heavily heavy hebdomadally hectically heedfully
heedlessly heinously hell-for-leather hellishly helpfully helplessly helter-
skelter hence henceforth henceforward here hereabout hereabouts hereafter hereby
herein hereinafter hereinbefore hereof hereto heretofore hereunder hereupon
herewith hermetically heroically hesitantly hesitatingly hideously
hierarchically hieroglyphically higgledy-piggledy high high-handedly high-
mindedly highly hilariously hinderingly histologically historically hither
hitherto hoarsely home homeostatically homeward homewards homogeneously honestly
honorably honourably hopefully hopelessly horizontally horribly horridly
horrifyingly horseback horticulturally hospitably hostilely hotfoot hotly hourly
however huffily hugely hugger-mugger humanely humanly humbly humiliatingly
humorlessly humorously humourlessly hundredfold hungrily hurriedly huskily
hydraulically hydraulicly hygienically hyperbolically hypnotically
hypocritically hypothalamically hypothetically hysterically i.e. ib. ibid.
ibidem icily ideally identically identifiably ideographically ideologically
idiomatically idiotically idly idolatrously idyllically ie ignobly ignominiously
ignorantly ill illegally illegibly illegitimately illiberally illicitly
illogically illustriously imaginatively immaculately immaturely immeasurably
immediately immensely imminently immoderately immodestly immorally immovably
immunologically immutably impalpably impartially impassively impatiently
impeccably impenitently imperatively imperceptibly imperfectly imperially
imperiously impermissibly impersonally impertinently impetuously impiously
impishly implausibly implicitly imploringly impolitely importantly importunately
imposingly impossibly impotently impracticably imprecisely impregnably
impressively improbably impromptu improperly improvidently imprudently
impudently impulsively in inaccessibly inaccurately inadequately inadvertently
inadvisably inalienably inanely inappropriately inarticulately inattentively
inaudibly inaugurally inauspiciously incautiously incessantly incestuously
incidentally incisively incognito incoherently incomparably incompatibly
incompetently incompletely inconceivably inconclusively incongruously
inconsequentially inconsequently inconsiderately inconsistently inconspicuously
incontrovertibly inconveniently incorrectly increasingly incredibly
incredulously incriminatingly incurably indecently indecisively indecorously
indeed indefatigably indefinitely indelibly independently indescribably
indeterminably indifferently indigenously indignantly indirectly indiscreetly
indiscriminately indistinctly individualistically individually indolently
indoors indubitably indulgently industrially industriously ineffably
ineffectively ineffectually inefficaciously inefficiently inelegantly
ineloquently ineluctably ineptly inequitably inescapably inevitably inexactly
inexcusably inexhaustibly inexorably inexpediently inexpensively inexpertly
inexpressively inextricably infectiously infelicitously infernally infinitely
inflexibly influentially informally informatively infra infrequently ingeniously
ingenuously ingloriously ingratiatingly inherently inhospitably inhumanely
inimitably iniquitously initially injudiciously injuriously inland innately
innocently inoffensively inopportunely inordinately inorganically inquiringly
inquisitively insanely insatiably inscriptively inscrutably insecticidally
insecurely insensately insensibly insensitively inseparably inshore inside
insidiously insignificantly insincerely insinuatingly insipidly insistently
insofar insolently insomuch inspirationally instantaneously instantly instead
instinctively institutionally instructively insubstantially insufficiently
insultingly insuperably integrally intellectually intelligently intelligibly
intemperately intensely intensively intentionally intently interchangeably
interdepartmental interestingly intermediately interminably intermittently
internally internationally interracially interrogatively intimately intolerably
intolerantly intractably intradermally intramuscularly intransitively
intravenously intrepidly intricately intrinsically intuitively invariably
inventively inversely inveterate invidiously invincibly invisibly invitingly
involuntarily inward inwardly inwards irately ironically irrationally
irregardless irregularly irrelevantly irreparably irreproachably irresistibly
irresolutely irrespective irresponsibly irretrievably irreverently irreversibly
irrevocably irritably irritatingly isotropically item jaggedly jarringly
jauntily jealously jeeringly jejunely jerkily jestingly jocosely jocular jointly
jokingly jolly journalistically jovially joyfully joylessly joyously jubilantly
judicially judiciously jurisprudentially just justifiably justifiedly justly
keenly killingly kinaesthetically kinda kindly kinesthetically knavishly knee-
deep knee-high knowingly laboriously lackadaisically laconically lamely
lamentably landward landwards langsyne languidly languorously large largely
largo lasciviously last lastingly lastly late lately later laterally latterly
laudably laughably laughingly lavishly lawfully lawlessly laxly lazily learnedly
least leastways leastwise leeward left legally legato legibly legislatively
legitimately leisurely lengthily lengthways lengthwise leniently lento less
lethargically lewdly lexically liberally licentiously licitly lief lifelessly
light light-handedly light-headedly light-heartedly lightly lightsomely likely
likewise limitedly limnologically limpidly limply lineally linearly lingeringly
lingually linguistically lispingly listlessly literally literatim little live
lividly locally loftily logarithmically logically logogrammatically long long-
windedly longer longest longingly longitudinally longways longwise loose loosely
lopsidedly loquaciously lots loud loudly lovingly low loweringly lowest loyally
lucidly luckily ludicrously lugubriously lukewarmly luridly lusciously lustfully
lustily luxuriantly luxuriously lyrically macroscopically madly magically
magisterially magnanimously magnetically magnificently magniloquently mainly
majestically maladroitly malapropos malevolently maliciously malignantly
malignly man-to-man manageably managerially mandatorily manfully mangily
maniacally manifestly manipulatively manly manually marginally markedly
martially marvellously marvelously masochistically massively masterfully
materialistically materially maternally mathematically matrilineally maturely
mawkishly maximally maybe mayhap meagerly meagrely meanderingly meaningfully
meanly meanspiritedly meantime meanwhile measurably measuredly mechanically
mechanistically medially medically medicinally meditatively meekly mellow
mellowingly mellowly melodically melodiously melodramatically memorably
menacingly mendaciously menially mentally mercifully mercilessly merely
meretriciously meritoriously merrily messily metabolically metaphorically
metaphysically meteorologically methodically methodologically meticulously
metonymically metrically microscopically middling midmost midships midway
midweek mightily mighty mildly militarily millionfold mincingly mindfully
mindlessly minimally ministerially minutely miraculously mirthfully
mischievously miserably misleadingly mistakenly mistily mistrustfully mockingly
moderately modestly modishly moistly molto momentarily momently momentously
monaurally monolingually monosyllabically monotonously monstrously monthly
moodily moonily morally morbidly mordaciously more moreover morosely
morphologically mortally most mostly motherly motionlessly mournfully movingly
much mulishly multifariously multilaterally multiplicatively multiply mundanely
municipally munificently murderously murkily musically musicologically musingly
mutely mutually mysteriously mystically naively nakedly namely narrow-mindedly
narrowly nasally nastily nationally nationwide nattily naturally naughtily nay
ne'er near nearby nearer nearest nearly neatly nebulously necessarily needfully
needlessly needs nefariously negatively neglectfully negligently nervelessly
nervily nervously neurobiological neurotically never nevermore nevertheless new
newly next nicely nigh nigher nighest nightly nimbly ninefold no nobly
nocturnally nohow noiselessly noisily nominally non non-verbally nonchalantly
noncompetitively noncomprehensively none nonetheless nonlexically
nonspecifically nonstop nonverbally nonviolently nor'-east nor'-nor'-east
nor'-nor'-west nor'-west normally north north-east north-northeast north-
northwest north-west northeast northeastward northeastwardly northerly northward
northwards northwest northwestward northwestwardly nostalgically not notably
nothing noticeably notoriously notwithstanding now nowadays nowhere nowise
noxiously numbly numerically nutritionally nuttily o'clock o'er o.k. obdurately
obediently objectionably objectively obligatorily obligingly obliquely
obnoxiously obscenely obscurely obsequiously observably observantly observingly
obsessionally obsessively obstinately obstreperously obstructively obtrusively
obtusely obviously occasionally oddly odiously off off-hand off-the-clock
offensively offhand offhanded offhandedly officially officiously offshore
offside offstage oft often oftener oftentimes ofttimes ok okay ominously on once
one-on-one one-sidedly onerously only onshore onstage onward onwards opaquely
openly operationally operatively opportunely opposite oppositely oppressively
optically optimally optimistically optionally opulently orad orally ordinarily
organically organizationally originally ornamentally ornately osmotically
ostensibly ostentatiously otherwise out outdoors outlandishly outrageously
outright outside outspokenly outstandingly outward outwardly outwards over
overbearingly overboard overhead overleaf overly overmuch overnight
overpoweringly oversea overseas overside overtime overtly overwhelmingly
owlishly p.a. p.m. pacifically pacifistically painfully painlessly painstakingly
palatably palely pallidly palmately palpably paradoxically parasitically
pardonably parentally parenterally parenthetically parochially part part-time
partially particularly partly passably passim passing passionately passively
past pat patchily patently paternally pathetically pathogenically pathologically
patiently patrilineally patriotically patronisingly patronizingly peaceably
peacefully peculiarly pedagogically pedantically peevishly pejoratively pell-
mell pellucidly penally penetratingly penetratively penitentially penitently
pensively penuriously peradventure perceptibly perceptively perceptually
perchance peremptorily perennially perfectly perfidiously perforce perfunctorily
perhaps perilously periodically peripherally perkily permanently permissibly
permissively perniciously perpendicularly perpetually perplexedly perseveringly
persistently person-to-person personally perspicuously persuasively
pertinaciously pertinently pertly pervasively perversely pessimistically pettily
pettishly petulantly pharmacologically phenomenally philanthropically
philatelically philosophically phlegmatically phonemic phonetically
photoelectrically photographically photometrically phylogenetically physically
physiologically pianissimo piano pickaback pictorially picturesquely piecemeal
piercingly pig-a-back pig-headedly piggishly piggyback pinnately piously piping
piquantly piratically pit-a-pat piteously pithily pitiably pitifully pitilessly
pitter-patter pitty-pat pitty-patty pityingly pizzicato placatingly placidly
plaguey plaguily plaguy plain plainly plaintively plastically plausibly
playfully pleadingly pleasantly please pleasingly pleasurably plenarily
plenteously plentifully plenty ploddingly plop pluckily plum plumb plump plunk
pneumatically poetically poignantly point-blank pointedly pointlessly
poisonously polemically politely politically polygonally polyphonically
polysyllabically pompously ponderously poorly pop popishly popularly
pornographically portentously positively possessively possibly post-free post-
haste post-paid posthumously postoperatively potentially potently poutingly
powerful powerfully powerlessly practicably practically pragmatically
praiseworthily pre-eminently precariously precedentedly precious preciously
precipitately precipitously precisely precociously predicatively predictably
predominantly preeminently preferably preferentially prematurely preponderantly
prepositionally preposterously presciently presentably presently presidentially
pressingly prestissimo presto presumably presumptively presumptuously
pretentiously preternaturally prettily pretty previously priggishly primarily
primitively primly principally prissily privately privily prn pro
probabilistically probably problematically prodigally prodigiously productively
profanely professedly professionally professorially proficiently profitably
profitlessly profligately profoundly profusely progressively prohibitively
prominently promiscuously promisingly promptly pronto properly prophetically
propitiously proportionally proportionately prosaically prosily prosperously
protectively protractedly proudly provably proverbially providentially
providently provincially provisionally provocatively provokingly prudently
prudishly pruriently pryingly psychically psychologically publically publicly
puckishly pugnaciously punctiliously punctually pungently punily punishingly
punitively punitorily purely puritanically purportedly purposefully
purposelessly purposely pusillanimously pyramidically quaintly qualitatively
quantitatively quarterly quaveringly queasily queerly querulously questionably
questioningly quick quicker quickest quickly quiet quietly quite quixotically
quizzically racially racily radially radiantly radically radioactively raffishly
raggedly rakishly rallentando ramblingly rampantly randomly rapaciously rapidly
rapturously rarely rashly raspingly rather rationally rattling raucously
ravenously raving ravingly ravishingly readably readily real realistically
really rearward rearwards reasonably reassuringly rebelliously rebukingly
recently receptively reciprocally recklessly recognizably recurrently red-handed
redly reflectively reflexly refreshfully refreshingly regally regardless
regimentally regionally regretfully regrettably regularly relatively
relativistically relentlessly relevantly reliably religiously reluctantly
remarkably reminiscently remorsefully remorselessly remotely rent-free
repeatedly repellently repellingly repentantly repetitively reportedly
reprehensibly reprehensively reproachfully reproducibly reprovingly repulsively
reputably reputedly resentfully reservedly residentially resignedly resolutely
resoundingly resourcefully respectably respectfully respectively resplendently
responsibly restfully restively restlessly restrictively retail retentively
reticently retroactively retrospectively revengefully reverentially reverently
reversely reversibly revoltingly rewardingly rhapsodically rhetorically
rhythmically richly ridiculously right right-down righteously rightfully rightly
rigidly rigorously riotously ripely riskily ritually roaring robustly roguishly
rollickingly romantically roomily rotationally rottenly rotundly rough roughly
round round-arm roundly routinely rowdily royally rudely ruefully ruggedly
ruinously rurally ruthlessly sacredly sacrilegiously sadly safely sagaciously
sagely salaciously sanctimoniously sanely sapiently sarcastically sardonically
satirically satisfactorily satisfyingly saucily savagely scandalously scantily
scarce scarcely scarily scathingly scenically sceptically schematically
schismatically scholastically scienter scientifically scoffingly scorching
scornfully scrappily screakily screamingly scrumptiously scrupulously
scurrilously scurvily searchingly seasonably seasonally seaward seawards second
second-best secondarily secondhand secondly secretively secretly securely
sedately seductively sedulously seemingly seldom selectively self-conceitedly
self-consciously self-evidently self-indulgently self-righteously selfishly
selflessly semantically semiannually semimonthly semiweekly sensationally
senselessly sensibly sensitively sensually sensuously sententiously
sentimentally separably separately sequentially serenely serially seriatim
seriously servilely sevenfold seventhly severally severely sexually shabbily
shaggily shakily shallowly shambolically shamefacedly shamefully shamelessly
shapelessly sharp sharply sheepishly sheer shiftily shockingly shoddily short
shortly shoulder-to-shoulder showily shrewdly shrewishly shrilly shudderingly
shyly sic sickeningly sidearm sidelong sidesaddle sidesplittingly sideward
sidewards sideway sideways sidewise signally significantly silently silkily
similarly simperingly simply simultaneously sincerely single-handed single-
handedly single-mindedly singly singularly sinuously sinusoidally sixfold
sixthly skeptically sketchily skew-whiff skilfully skillfully skimpily
skittishly sky-high skyward skywards slackly slam-bang slanderously slangily
slantingly slantways slantwise slap slap-bang slapdash slavishly sleekly
sleepily sleeplessly slenderly slickly slightingly slightly slimly slopingly
sloppily slouchily slouchingly slow slower slowest slowly sluggishly slyly smack
small small-mindedly smarmily smartly smash smashingly smilingly smolderingly
smoothly smoulderingly smugly smuttily snappishly sneakily sneakingly sneeringly
snidely snobbishly snootily snugly so so-so soaking sobbingly soberly sociably
socially sociobiologically socioeconomically sociolinguistically sociologically
soft softly solely solemnly solicitously solidly solitarily solo somberly
sombrely some someday somehow someplace sometime sometimes someway someways
somewhat somewhere somnolently sonorously soon sooner soonest soothingly sopping
sordidly sorely sorrowfully sottishly sou'-east sou'-sou'-east sou'-sou'-west
sou'west soughingly soulfully soullessly soundlessly soundly sourly south south-
east south-southeast south-southwest south-west southeast southeastward
southeastwardly southerly southward southwards southwest southwestward
southwestwardly spaceward spacewards spaciously sparely sparingly sparsely
spasmodically spatially specially specifically speciously spectacularly
spectrographically speculatively speechlessly speedily spherically spicily
spinally spirally spiritedly spiritually spitefully splendidly spontaneously
spookily sporadically sportingly sportively spotlessly sprucely spuriously
squalidly square squarely squeamishly stably staccato staggeringly stagily
staidly stanchly standoffishly stark starkly startlingly statistically
statutorily staunchly steadfastly steadily steady stealthily steaming steeply
stepwise stereotypically sternly stertorously stickily stiff stiffly still
stiltedly stingily stirringly stochastically stock-still stockily stodgily
stoically stolidly stonily stormily stoutly stragglingly straight straight-
backed straightaway straightforwardly straightway strangely strategically
strenuously strictly stridently strikingly stringently strongly structurally
stubbornly studiously stuffily stunningly stupendously stupidly sturdily
stylishly stylistically suavely subconsciously subcutaneously subjectively
sublimely submissively subsequently subserviently substantially subtly
successfully successively succinctly such suddenly sufficiently suggestively
suitably sulkily sullenly sultrily summarily sumptuously sunnily super superbly
superciliously superficially superfluously superlatively supernaturally
superstitiously supinely supposedly supra supremely sure surely surgically
surlily surpassingly surprisedly surprisingly surreptitiously suspiciously
sweepingly sweet sweetly swiftly swimmingly syllabically symbiotically
symbolically symmetrically sympathetically symptomatically synchronously
synergistically synonymously syntactically synthetically systematically tacitly
taciturnly tactfully tactically tactlessly tactually talkatively talkily tamely
tandem tangentially tangibly tantalizingly tardily tartly tastefully tastelessly
tastily tattily tauntingly tautly tawdrily taxonomically tearfully teasingly
technically technologically tediously telegraphically telescopically tellingly
temperamentally temperately temporally temporarily temptingly tenaciously
tendentiously tenderly tenfold tensely tentatively tenthly tenuously tepidly
terminally terrestrially terribly terrifically territorially tersely testily
tetchily thankfully theatrically thematically then thence thenceforth
theologically theoretically therapeutically there thereabout thereabouts
thereafter thereby therefor therefore therefrom therein thereinafter thereof
thereon thereto theretofore thereunder therewith therewithal thermally
thermodynamically thermostatically thick thickly thievishly thin thinly third
thirdhand thirdly thirstily thither thoroughly though thoughtfully thoughtlessly
thousand-fold threateningly threefold thrice thriftily thriftlessly through
throughout thus thusly tidily tight tightly timely timidly timorously tip-top
tiptoe tiredly tirelessly tiresomely today toe-to-toe together tolerably
tolerantly tomorrow tonelessly tongue-in-cheek tonight too topically
topographically topologically toppingly topsy-turvily topsy-turvy torpidly
tortuously torturously totally touchily touchingly toughly traditionally
tragically traitorously tranquilly transcendentally transiently transitionally
transitively transitorily transparently transversally transversely treacherously
treasonably tremendously tremulously trenchantly trepidly trickily trimly
trippingly tritely triumphantly trivially tropically truculently true truly
trustfully trustingly truthfully tumultuously tunefully tunelessly turbulently
turgidly tutorially twice twirlingly twofold typically typographically
ulteriorly ultimately ultrasonically unabashedly unacceptably unaccompanied
unaccountably unachievably unadvisedly unalterably unambiguously unambitiously
unanimously unappealingly unappreciatively unarguably unashamedly unassailably
unassertively unassumingly unattainably unattractively unavoidably unawares
unbearably unbecomingly unbeknown unbeknownst unbelievably unbelievingly
unblinkingly unblushingly uncannily unceasingly unceremoniously uncertainly
unchangeably uncharacteristically unchivalrously uncivilly unclearly unco
uncomfortably uncommonly uncomparably uncomplainingly uncompromisingly
unconcernedly unconditionally unconsciously unconstitutionally uncontrollably
uncontroversially unconventionally unconvincingly uncouthly uncritically
unctuously undecipherably undemocratically undeniably undependably under
underarm underfoot underground underhand underhandedly underneath understandably
understandingly undeservedly undesirably undiplomatically undisputedly
undoubtedly undramatically unduly uneasily unemotionally unendingly
unenergetically unenthusiastically unequally unequivocally unerringly
unethically unevenly uneventfully unexcitingly unexpectedly unfailingly unfairly
unfaithfully unfalteringly unfashionably unfavorably unfavourably unfeelingly
unfeignedly unforgettably unforgivably unforgivingly unfortunately ungracefully
ungraciously ungrammatically ungratefully ungrudgingly unhappily unharmoniously
unhelpfully unhesitatingly unhurriedly unhygienically uniformly unilaterally
unimaginably unimaginatively unimpeachably unimpressively uninformatively
uninstructively unintelligently unintelligibly unintentionally uninterestingly
uninterruptedly uninvitedly uniquely unitedly universally unjustifiably unjustly
unkindly unknowingly unlawfully unluckily unmanageably unmanfully unmanly
unmelodiously unmemorably unmercifully unmindfully unmistakably unmusically
unnaturally unnecessarily unnoticeably unobtrusively unofficially unoriginally
unpalatably unpardonably unpatriotically unpleasantly unprecedentedly
unpredictably unpretentiously unproductively unprofitably unpropitiously
unqualifiedly unquestionably unquestioningly unquietly unreadably
unrealistically unreasonably unreasoningly unrecognisable unrecognizably
unrelentingly unreliably unremarkably unrepentantly unreproducibly unreservedly
unrestrainedly unrighteously unromantically unsatiably unsatisfactorily
unscientifically unscrupulously unseasonably unselfconsciously unselfishly
unsentimentally unshakably unsmilingly unsociably unsparingly unspeakably
unsportingly unsteadily unstintingly unsuccessfully unsuitably unsuspectingly
unswervingly unsymmetrically unsympathetically unsystematically unthinkably
unthinking unthinkingly untidily untimely untruly untruthfully untypically
ununderstandably unusually unutterably unwantedly unwarily unwarrantably
unwaveringly unwillingly unwisely unwittingly unwontedly unworthily up up-
country uphill uppermost uppishly uprightly upriver uproariously upstage
upstairs upstate upstream uptown upward upwardly upwards upwind urbanely
urgently usefully uselessly usually usuriously utterly uxoriously vacantly
vacuously vaguely vainly valiantly validly valorously vanishingly vapidly
variably variously vastly vauntingly vehemently venally vengefully venomously
ventrally verbally verbatim verbosely verily vertically very vexatiously
vicariously viciously victoriously videlicet vigilantly vigorously vilely
vindictively violently virtually virtuously virulently vis-a-vis viscerally
viscidly visibly visually vitally vitriolically vivace vivaciously vividly viz.
vocally vocationally vociferously volcanically volitionally volubly
volumetrically voluntarily voluptuously voraciously voyeuristically vulgarly
vulnerably wackily wafer-thin waggishly waist-deep waist-high wanly wantonly
warily warm warmly wastefully watchfully way weakly wealthily wearily weekly
weightily weirdly well well-nigh well-timed west westerly westward westwards
whacking wheezily wheezingly whence wheresoever wherever whimsically whole
wholeheartedly wholesale wholesomely wholly whopping wickedly wide widely wild
wildly wilfully willfully willingly willy-nilly windily windward winsomely
wisely wishfully wistfully withal witheringly within wittily wittingly woefully
wolfishly wonderfully wonderingly wondrous wondrously woodenly wordily
wordlessly worriedly worryingly worse worst worthily worthlessly wrathfully
wretchedly wrong wrongfully wrongheadedly wrongly wryly yea yeah yearly
yearningly yesterday yet yieldingly yon yonder youthfully zealously zestfully
zestily zigzag
""".split())



# coding: utf8
"""Based on unique combinations form this dataset:

https://github.com/UniversalDependencies/UD_Norwegian-Bokmaal

Models can be trained using this. dataset as well."""
from __future__ import unicode_literals

from ...symbols import POS, PUNCT, ADJ, CONJ, CCONJ, SCONJ, SYM, NUM, DET, ADV, ADP, X, VERB
from ...symbols import NOUN, PROPN, PART, INTJ, SPACE, PRON, AUX

TAG_MAP = {
    'NOUN__Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing', POS: NOUN},
    'CCONJ___': {'morph': '_', POS: CCONJ},
    'NOUN__Definite=Ind|Gender=Masc|Number=Sing': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing', POS: NOUN},
    'ADP___': {'morph': '_', POS: ADP},
    'NOUN__Definite=Def|Gender=Masc|Number=Sing': {'morph': 'Definite=Def|Gender=Masc|Number=Sing', POS: NOUN},
    'NOUN__Definite=Ind|Gender=Neut|Number=Plur': {'morph': 'Definite=Ind|Gender=Neut|Number=Plur', POS: NOUN},
    'PUNCT___': {'morph': '_', POS: PUNCT},
    'VERB__Mood=Ind|Tense=Past|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Past|VerbForm=Fin', POS: VERB},
    'DET__Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'ADJ__Definite=Def|Degree=Pos|Number=Sing': {'morph': 'Definite=Def|Degree=Pos|Number=Sing', POS: ADJ},
    'PROPN___': {'morph': '_', POS: PROPN},
    'X___': {'morph': '_', POS: X},
    'VERB__Mood=Ind|Tense=Pres|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin', POS: VERB},
    'NOUN__Definite=Def|Gender=Neut|Number=Sing': {'morph': 'Definite=Def|Gender=Neut|Number=Sing', POS: NOUN},
    'PRON__PronType=Rel': {'morph': 'PronType=Rel', POS: PRON},
    'AUX__Mood=Ind|Tense=Pres|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin', POS: AUX},
    'ADJ__Definite=Ind|Gender=Neut|Number=Sing|VerbForm=Part': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing|VerbForm=Part', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Pos|Number=Sing': {'morph': 'Definite=Ind|Degree=Pos|Number=Sing', POS: ADJ},
    'NOUN__Definite=Ind|Gender=Fem|Number=Sing': {'morph': 'Definite=Ind|Gender=Fem|Number=Sing', POS: NOUN},
    'ADJ__Number=Plur|VerbForm=Part': {'morph': 'Number=Plur|VerbForm=Part', POS: ADJ},
    'NOUN__Definite=Ind|Gender=Fem|Number=Plur': {'morph': 'Definite=Ind|Gender=Fem|Number=Plur', POS: NOUN},
    'ADV___': {'morph': '_', POS: ADV},
    'PRON__Gender=Neut|Number=Sing|Person=3|PronType=Prs': {'morph': 'Gender=Neut|Number=Sing|Person=3|PronType=Prs', POS: PRON},
    'ADJ__Definite=Ind|Number=Sing|VerbForm=Part': {'morph': 'Definite=Ind|Number=Sing|VerbForm=Part', POS: ADJ},
    'VERB__VerbForm=Part': {'morph': 'VerbForm=Part', POS: VERB},
    'NOUN__Definite=Ind|Gender=Masc|Number=Plur': {'morph': 'Definite=Ind|Gender=Masc|Number=Plur', POS: NOUN},
    'ADJ__Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing', POS: ADJ},
    'ADJ__Degree=Pos|Number=Plur': {'morph': 'Degree=Pos|Number=Plur', POS: ADJ},
    'NUM__Number=Plur|NumType=Card': {'morph': 'Number=Plur|NumType=Card', POS: NUM},
    'NOUN__Definite=Def|Gender=Masc|Number=Plur': {'morph': 'Definite=Def|Gender=Masc|Number=Plur', POS: NOUN},
    'PRON__Case=Acc|PronType=Prs|Reflex=Yes': {'morph': 'Case=Acc|PronType=Prs|Reflex=Yes', POS: PRON},
    'NOUN__Case=Gen|Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Gender=Neut|Number=Sing', POS: NOUN},
    'PART___': {'morph': '_', POS: PART},
    'VERB__VerbForm=Inf': {'morph': 'VerbForm=Inf', POS: VERB},
    'PRON__Case=Nom|Number=Plur|Person=3|PronType=Prs': {'morph': 'Case=Nom|Number=Plur|Person=3|PronType=Prs', POS: PRON},
    'AUX__Mood=Ind|Tense=Past|VerbForm=Fin': {'morph': 'Mood=Ind|Tense=Past|VerbForm=Fin', POS: AUX},
    'PROPN__Gender=Fem': {'morph': 'Gender=Fem', POS: PROPN},
    'NOUN___': {'morph': '_', POS: NOUN},
    'PROPN__Gender=Masc': {'morph': 'Gender=Masc', POS: PROPN},
    'DET__Gender=Neut|Number=Sing|PronType=Dem': {'morph': 'Gender=Neut|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Masc|Number=Sing|PronType=Art': {'morph': 'Gender=Masc|Number=Sing|PronType=Art', POS: DET},
    'NOUN__Case=Gen|Definite=Def|Gender=Masc|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Gender=Masc|Number=Sing', POS: NOUN},
    'PROPN__Abbr=Yes': {'morph': 'Abbr=Yes', POS: PROPN},
    'ADV__Polarity=Neg': {'morph': 'Polarity=Neg', POS: ADV},
    'SCONJ___': {'morph': '_', POS: SCONJ},
    'DET__Number=Plur|Poss=Yes|PronType=Prs': {'morph': 'Number=Plur|Poss=Yes|PronType=Prs', POS: DET},
    'NOUN__Case=Gen|Definite=Ind|Gender=Neut|Number=Plur': {'morph': 'Case=Gen|Definite=Ind|Gender=Neut|Number=Plur', POS: NOUN},
    'PROPN__Case=Gen': {'morph': 'Case=Gen', POS: PROPN},
    'DET__Gender=Fem|Number=Sing|PronType=Dem': {'morph': 'Gender=Fem|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Masc|Number=Sing|Poss=Yes|PronType=Prs': {'morph': 'Gender=Masc|Number=Sing|Poss=Yes|PronType=Prs', POS: DET},
    'ADJ__Definite=Def|Degree=Sup': {'morph': 'Definite=Def|Degree=Sup', POS: ADJ},
    'PROPN__Case=Gen|Gender=Fem': {'morph': 'Case=Gen|Gender=Fem', POS: PROPN},
    'DET__Number=Plur|PronType=Dem': {'morph': 'Number=Plur|PronType=Dem', POS: DET},
    'NOUN__Case=Gen|Definite=Def|Gender=Neut|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Gender=Neut|Number=Sing', POS: NOUN},
    'ADJ__Definite=Ind|Degree=Sup': {'morph': 'Definite=Ind|Degree=Sup', POS: ADJ},
    'NOUN__Definite=Def|Gender=Fem|Number=Plur': {'morph': 'Definite=Def|Gender=Fem|Number=Plur', POS: NOUN},
    'PROPN__Gender=Neut': {'morph': 'Gender=Neut', POS: PROPN},
    'DET__Number=Plur|PronType=Int': {'morph': 'Number=Plur|PronType=Int', POS: DET},
    'NOUN__Definite=Def|Gender=Neut|Number=Plur': {'morph': 'Definite=Def|Gender=Neut|Number=Plur', POS: NOUN},
    'DET__Definite=Def|PronType=Dem': {'morph': 'Definite=Def|PronType=Dem', POS: DET},
    'DET__Gender=Neut|Number=Sing|PronType=Art': {'morph': 'Gender=Neut|Number=Sing|PronType=Art', POS: DET},
    'VERB__Mood=Ind|Tense=Pres|VerbForm=Fin|Voice=Pass': {'morph': 'Mood=Ind|Tense=Pres|VerbForm=Fin|Voice=Pass', POS: VERB},
    'PROPN__Abbr=Yes|Case=Gen': {'morph': 'Abbr=Yes|Case=Gen', POS: PROPN},
    'PRON__Animacy=Hum|Case=Nom|Gender=Masc|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Gender=Masc|Number=Sing|Person=3|PronType=Prs', POS: PRON},
    'ADJ__Degree=Cmp': {'morph': 'Degree=Cmp', POS: ADJ},
    'ADJ__VerbForm=Part': {'morph': 'VerbForm=Part', POS: ADJ},
    'DET__Gender=Neut|Number=Sing|Poss=Yes|PronType=Prs': {'morph': 'Gender=Neut|Number=Sing|Poss=Yes|PronType=Prs', POS: DET},
    'ADP__Abbr=Yes': {'morph': 'Abbr=Yes', POS: ADP},
    'DET__Definite=Ind|Gender=Neut|Number=Sing|PronType=Prs': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing|PronType=Prs', POS: DET},
    'NOUN__Case=Gen|Definite=Def|Gender=Neut|Number=Plur': {'morph': 'Case=Gen|Definite=Def|Gender=Neut|Number=Plur', POS: NOUN},
    'AUX__VerbForm=Part': {'morph': 'VerbForm=Part', POS: AUX},
    'PRON__PronType=Int': {'morph': 'PronType=Int', POS: PRON},
    'DET__Gender=Fem|Number=Sing|Poss=Yes|PronType=Prs': {'morph': 'Gender=Fem|Number=Sing|Poss=Yes|PronType=Prs', POS: DET},
    'PRON__Number=Plur|Person=3|PronType=Ind,Prs': {'morph': 'Number=Plur|Person=3|PronType=Ind,Prs', POS: PRON},
    'DET__Number=Plur|PronType=Ind': {'morph': 'Number=Plur|PronType=Ind', POS: DET},
    'ADJ__Degree=Pos': {'morph': 'Degree=Pos', POS: ADJ},
    'PRON__Animacy=Hum|Case=Nom|Number=Plur|Person=1|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Number=Plur|Person=1|PronType=Prs', POS: PRON},
    'VERB__VerbForm=Inf|Voice=Pass': {'morph': 'VerbForm=Inf|Voice=Pass', POS: VERB},
    'DET__Definite=Ind|Gender=Fem|Number=Sing|PronType=Dem': {'morph': 'Definite=Ind|Gender=Fem|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Neut|Number=Sing|PronType=Ind': {'morph': 'Gender=Neut|Number=Sing|PronType=Ind', POS: DET},
    'PRON__Animacy=Hum|Case=Acc|Gender=Masc|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Hum|Case=Acc|Gender=Masc|Number=Sing|Person=3|PronType=Prs', POS: PRON},
    'PRON__Animacy=Hum|Case=Nom|Number=Sing|Person=1|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Number=Sing|Person=1|PronType=Prs', POS: PRON},
    'DET__Number=Plur|Polarity=Neg|PronType=Neg': {'morph': 'Number=Plur|Polarity=Neg|PronType=Neg', POS: DET},
    'NUM__NumType=Card': {'morph': 'NumType=Card', POS: NUM},
    'DET__Gender=Masc|Number=Sing|PronType=Ind': {'morph': 'Gender=Masc|Number=Sing|PronType=Ind', POS: DET},
    'DET__PronType=Prs': {'morph': 'PronType=Prs', POS: DET},
    'DET__Gender=Fem|Number=Sing|PronType=Ind': {'morph': 'Gender=Fem|Number=Sing|PronType=Ind', POS: DET},
    'PROPN__Case=Gen|Gender=Neut': {'morph': 'Case=Gen|Gender=Neut', POS: PROPN},
    'DET__Gender=Masc|Number=Sing|Polarity=Neg|PronType=Neg': {'morph': 'Gender=Masc|Number=Sing|Polarity=Neg|PronType=Neg', POS: DET},
    'ADJ__Definite=Def|Number=Sing|VerbForm=Part': {'morph': 'Definite=Def|Number=Sing|VerbForm=Part', POS: ADJ},
    'PRON__Gender=Fem,Masc|Number=Sing|Person=3|PronType=Prs': {'morph': 'Gender=Fem,Masc|Number=Sing|Person=3|PronType=Prs', POS: PRON},
    'AUX__VerbForm=Inf': {'morph': 'VerbForm=Inf', POS: AUX},
    'PRON__Case=Acc|Number=Plur|Person=3|PronType=Prs': {'morph': 'Case=Acc|Number=Plur|Person=3|PronType=Prs', POS: PRON},
    'ADJ__Case=Gen|Degree=Pos|Number=Plur': {'morph': 'Case=Gen|Degree=Pos|Number=Plur', POS: ADJ},
    'DET__Number=Plur|PronType=Tot': {'morph': 'Number=Plur|PronType=Tot', POS: DET},
    'DET__Case=Gen|Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'DET__Number=Plur|PronType=Prs': {'morph': 'Number=Plur|PronType=Prs', POS: DET},
    'SYM___': {'morph': '_', POS: SYM},
    'NUM__Gender=Neut|Number=Sing|NumType=Card': {'morph': 'Gender=Neut|Number=Sing|NumType=Card', POS: NUM},
    'PRON__Animacy=Hum|Case=Nom|Number=Sing|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Number=Sing|PronType=Prs', POS: PRON},
    'DET__Definite=Ind|Gender=Masc|Number=Sing|PronType=Prs': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing|PronType=Prs', POS: DET},
    'NOUN__Case=Gen|Definite=Ind|Gender=Masc|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Gender=Masc|Number=Sing', POS: NOUN},
    'ADV__Abbr=Yes': {'morph': 'Abbr=Yes', POS: ADV},
    'DET__Definite=Ind|Gender=Neut|Number=Sing|PronType=Dem': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Masc|Number=Sing|PronType=Tot': {'morph': 'Gender=Masc|Number=Sing|PronType=Tot', POS: DET},
    'DET__Definite=Def|PronType=Prs': {'morph': 'Definite=Def|PronType=Prs', POS: DET},
    'PRON__Animacy=Hum|Case=Nom|Gender=Fem|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Gender=Fem|Number=Sing|Person=3|PronType=Prs', POS: PRON},
    'NOUN__Gender=Neut': {'morph': 'Gender=Neut', POS: NOUN},
    'DET__Gender=Neut|Number=Sing|PronType=Int': {'morph': 'Gender=Neut|Number=Sing|PronType=Int', POS: DET},
    'NUM__Definite=Def|NumType=Card': {'morph': 'Definite=Def|NumType=Card', POS: NUM},
    'VERB__Mood=Imp|VerbForm=Fin': {'morph': 'Mood=Imp|VerbForm=Fin', POS: VERB},
    'NOUN__Definite=Ind|Number=Plur': {'morph': 'Definite=Ind|Number=Plur', POS: NOUN},
    'DET__Gender=Neut|Number=Sing|PronType=Tot': {'morph': 'Gender=Neut|Number=Sing|PronType=Tot', POS: DET},
    'DET__Gender=Fem|Number=Sing|PronType=Tot': {'morph': 'Gender=Fem|Number=Sing|PronType=Tot', POS: DET},
    'PRON__Animacy=Hum|Case=Acc|Number=Plur|Person=1|PronType=Prs': {'morph': 'Animacy=Hum|Case=Acc|Number=Plur|Person=1|PronType=Prs', POS: PRON},
    'PRON__Gender=Fem,Masc|Number=Sing|Person=3|Polarity=Neg|PronType=Neg,Prs': {'morph': 'Gender=Fem,Masc|Number=Sing|Person=3|Polarity=Neg|PronType=Neg,Prs', POS: PRON},
    'PRON__Number=Plur|Person=3|Polarity=Neg|PronType=Neg,Prs': {'morph': 'Number=Plur|Person=3|Polarity=Neg|PronType=Neg,Prs', POS: PRON},
    'NUM__Definite=Def|Number=Sing|NumType=Card': {'morph': 'Definite=Def|Number=Sing|NumType=Card', POS: NUM},
    'NUM__Gender=Masc|Number=Sing|NumType=Card': {'morph': 'Gender=Masc|Number=Sing|NumType=Card', POS: NUM},
    'DET__Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'NOUN__Case=Gen|Definite=Def|Gender=Fem|Number=Plur': {'morph': 'Case=Gen|Definite=Def|Gender=Fem|Number=Plur', POS: NOUN},
    'DET__Case=Gen|Gender=Neut|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Gender=Neut|Number=Sing|PronType=Dem', POS: DET},
    'PRON__Animacy=Hum|Number=Sing|PronType=Art,Prs': {'morph': 'Animacy=Hum|Number=Sing|PronType=Art,Prs', POS: PRON},
    'AUX__Mood=Imp|VerbForm=Fin': {'morph': 'Mood=Imp|VerbForm=Fin', POS: AUX},
    'PRON__Number=Plur|Person=3|PronType=Prs,Tot': {'morph': 'Number=Plur|Person=3|PronType=Prs,Tot', POS: PRON},
    'ADJ__Number=Plur': {'morph': 'Number=Plur', POS: ADJ},
    'NOUN__Gender=Masc': {'morph': 'Gender=Masc', POS: NOUN},
    'NOUN__Abbr=Yes': {'morph': 'Abbr=Yes', POS: NOUN},
    'NOUN__Case=Gen|Definite=Ind|Gender=Masc|Number=Plur': {'morph': 'Case=Gen|Definite=Ind|Gender=Masc|Number=Plur', POS: NOUN},
    'PRON__Gender=Neut|Number=Sing|Person=3|PronType=Ind,Prs': {'morph': 'Gender=Neut|Number=Sing|Person=3|PronType=Ind,Prs', POS: PRON},
    'INTJ___': {'morph': '_', POS: INTJ},
    'PRON__Animacy=Hum|Case=Nom|Number=Sing|Person=2|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Number=Sing|Person=2|PronType=Prs', POS: PRON},
    'PRON__Animacy=Hum|Case=Acc|Number=Sing|Person=1|PronType=Prs': {'morph': 'Animacy=Hum|Case=Acc|Number=Sing|Person=1|PronType=Prs', POS: PRON},
    'NOUN__Case=Gen|Definite=Def|Gender=Masc|Number=Plur': {'morph': 'Case=Gen|Definite=Def|Gender=Masc|Number=Plur', POS: NOUN},
    'ADJ___': {'morph': '_', POS: ADJ},
    'PRON__Animacy=Hum|Case=Acc|Gender=Fem|Number=Sing|Person=3|PronType=Prs': {'morph': 'Animacy=Hum|Case=Acc|Gender=Fem|Number=Sing|Person=3|PronType=Prs', POS: PRON},
    'PRON__Animacy=Hum|Case=Acc|Number=Sing|Person=2|PronType=Prs': {'morph': 'Animacy=Hum|Case=Acc|Number=Sing|Person=2|PronType=Prs', POS: PRON},
    'NOUN__Definite=Def|Gender=Fem|Number=Sing': {'morph': 'Definite=Def|Gender=Fem|Number=Sing', POS: NOUN},
    'PRON__Number=Sing|Polarity=Neg|PronType=Neg': {'morph': 'Number=Sing|Polarity=Neg|PronType=Neg', POS: PRON},
    'NOUN__Case=Gen': {'morph': 'Case=Gen', POS: NOUN},
    'ADJ__Definite=Ind|Number=Sing': {'morph': 'Definite=Ind|Number=Sing', POS: ADJ},
    'PROPN__Case=Gen|Gender=Masc': {'morph': 'Case=Gen|Gender=Masc', POS: PROPN},
    'PRON__Animacy=Hum|Number=Plur|PronType=Rcp': {'morph': 'Animacy=Hum|Number=Plur|PronType=Rcp', POS: PRON},
    'NOUN__Case=Gen|Definite=Ind|Gender=Fem|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Gender=Fem|Number=Sing', POS: NOUN},
    'PRON__Number=Plur|Person=3|PronType=Prs': {'morph': 'Number=Plur|Person=3|PronType=Prs', POS: PRON},
    'PRON__Gender=Fem,Masc|Number=Sing|Person=3|PronType=Ind,Prs': {'morph': 'Gender=Fem,Masc|Number=Sing|Person=3|PronType=Ind,Prs', POS: PRON},
    'DET__Definite=Ind|Gender=Fem|Number=Sing|PronType=Prs': {'morph': 'Definite=Ind|Gender=Fem|Number=Sing|PronType=Prs', POS: DET},
    'NOUN__Case=Gen|Definite=Def|Gender=Fem|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Gender=Fem|Number=Sing', POS: NOUN},
    'DET__Gender=Fem|Number=Sing|PronType=Art': {'morph': 'Gender=Fem|Number=Sing|PronType=Art', POS: DET},
    'ADJ__Case=Gen|Definite=Def|Degree=Pos|Number=Sing': {'morph': 'Case=Gen|Definite=Def|Degree=Pos|Number=Sing', POS: ADJ},
    'DET__Gender=Masc|Number=Sing|PronType=Int': {'morph': 'Gender=Masc|Number=Sing|PronType=Int', POS: DET},
    'NUM__Number=Sing|NumType=Card': {'morph': 'Number=Sing|NumType=Card', POS: NUM},
    'PRON__Animacy=Hum|Case=Acc|Number=Plur|Person=2|PronType=Prs': {'morph': 'Animacy=Hum|Case=Acc|Number=Plur|Person=2|PronType=Prs', POS: PRON},
    'PRON__Animacy=Hum|Case=Nom|Number=Plur|Person=2|PronType=Prs': {'morph': 'Animacy=Hum|Case=Nom|Number=Plur|Person=2|PronType=Prs', POS: PRON},
    'ADJ__Case=Gen|Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Degree=Pos|Gender=Neut|Number=Sing', POS: ADJ},
    'ADJ__Degree=Sup': {'morph': 'Degree=Sup', POS: ADJ},
    'PRON__Animacy=Hum|PronType=Int': {'morph': 'Animacy=Hum|PronType=Int', POS: PRON},
    'DET__PronType=Ind': {'morph': 'PronType=Ind', POS: DET},
    'DET__Definite=Def|Number=Sing|PronType=Dem': {'morph': 'Definite=Def|Number=Sing|PronType=Dem', POS: DET},
    'NOUN__Gender=Fem': {'morph': 'Gender=Fem', POS: NOUN},
    'DET__Case=Gen|Number=Plur|PronType=Dem': {'morph': 'Case=Gen|Number=Plur|PronType=Dem', POS: DET},
    'PRON__Gender=Fem,Masc|Number=Sing|Person=3|PronType=Prs,Tot': {'morph': 'Gender=Fem,Masc|Number=Sing|Person=3|PronType=Prs,Tot', POS: PRON},
    'NOUN__Case=Gen|Definite=Ind|Gender=Fem|Number=Plur': {'morph': 'Case=Gen|Definite=Ind|Gender=Fem|Number=Plur', POS: NOUN},
    'DET__Gender=Neut|Number=Sing|Polarity=Neg|PronType=Neg': {'morph': 'Gender=Neut|Number=Sing|Polarity=Neg|PronType=Neg', POS: DET},
    'NOUN__Number=Plur': {'morph': 'Number=Plur', POS: NOUN},
    'ADJ__Case=Gen|Definite=Ind|Degree=Pos|Number=Sing': {'morph': 'Case=Gen|Definite=Ind|Degree=Pos|Number=Sing', POS: ADJ},
    'VERB__Definite=Ind|Number=Sing|VerbForm=Part': {'morph': 'Definite=Ind|Number=Sing|VerbForm=Part', POS: VERB},
    'ADJ__Case=Gen|Definite=Def|Number=Sing|VerbForm=Part': {'morph': 'Case=Gen|Definite=Def|Number=Sing|VerbForm=Part', POS: ADJ},
    'VERB__Mood=Ind|Tense=Past|VerbForm=Fin|Voice=Pass': {'morph': 'Mood=Ind|Tense=Past|VerbForm=Fin|Voice=Pass', POS: VERB},
    'DET__Gender=Neut|Number=Sing|PronType=Dem,Ind': {'morph': 'Gender=Neut|Number=Sing|PronType=Dem,Ind', POS: DET},
    'PRON__Animacy=Hum|Poss=Yes|PronType=Int': {'morph': 'Animacy=Hum|Poss=Yes|PronType=Int', POS: PRON},
    'ADJ__Abbr=Yes': {'morph': 'Abbr=Yes', POS: ADJ},
    'DET__Case=Gen|Gender=Masc|Number=Sing|PronType=Art': {'morph': 'Case=Gen|Gender=Masc|Number=Sing|PronType=Art', POS: DET},
    'NOUN__Abbr=Yes|Definite=Def,Ind|Gender=Masc|Number=Sing': {'morph': 'Abbr=Yes|Definite=Def,Ind|Gender=Masc|Number=Sing', POS: NOUN},
    'PRON__PronType=Prs': {'morph': 'PronType=Prs', POS: PRON},
    'DET__Case=Gen|Gender=Fem|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Gender=Fem|Number=Sing|PronType=Dem', POS: DET},
    'DET__Number=Plur|Poss=Yes|PronType=Rcp': {'morph': 'Number=Plur|Poss=Yes|PronType=Rcp', POS: DET},
    'ADJ__Definite=Ind|Degree=Pos': {'morph': 'Definite=Ind|Degree=Pos', POS: ADJ},
    'DET__Number=Plur|PronType=Art': {'morph': 'Number=Plur|PronType=Art', POS: DET},
    'NUM__Case=Gen|Number=Plur|NumType=Card': {'morph': 'Case=Gen|Number=Plur|NumType=Card', POS: NUM},
    'NOUN__Abbr=Yes|Definite=Def,Ind|Gender=Neut|Number=Plur,Sing': {'morph': 'Abbr=Yes|Definite=Def,Ind|Gender=Neut|Number=Plur,Sing', POS: NOUN},
    'DET__Case=Gen|Number=Plur|PronType=Tot': {'morph': 'Case=Gen|Number=Plur|PronType=Tot', POS: DET},
    'NOUN__Abbr=Yes|Definite=Def,Ind|Gender=Masc|Number=Plur,Sing': {'morph': 'Abbr=Yes|Definite=Def,Ind|Gender=Masc|Number=Plur,Sing', POS: NOUN},
    'DET__Gender=Fem|Number=Sing|PronType=Int': {'morph': 'Gender=Fem|Number=Sing|PronType=Int', POS: DET},
    'ADJ__Definite=Ind|Gender=Neut|Number=Sing': {'morph': 'Definite=Ind|Gender=Neut|Number=Sing', POS: ADJ},
    'DET__Case=Gen|Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem': {'morph': 'Case=Gen|Definite=Ind|Gender=Masc|Number=Sing|PronType=Dem', POS: DET},
    'DET__Gender=Fem|Number=Sing|PronType=Prs': {'morph': 'Gender=Fem|Number=Sing|PronType=Prs', POS: DET},
    'PRON__Animacy=Hum|Case=Gen,Nom|Number=Sing|PronType=Art,Prs': {'morph': 'Animacy=Hum|Case=Gen,Nom|Number=Sing|PronType=Art,Prs', POS: PRON},
    'ADJ__Definite=Def|Degree=Pos|Gender=Masc|Number=Sing': {'morph': 'Definite=Def|Degree=Pos|Gender=Masc|Number=Sing', POS: ADJ},
    'PRON__Animacy=Hum|Case=Gen|Number=Sing|PronType=Art,Prs': {'morph': 'Animacy=Hum|Case=Gen|Number=Sing|PronType=Art,Prs', POS: PRON},
    'NUM__Gender=Fem|Number=Sing|NumType=Card': {'morph': 'Gender=Fem|Number=Sing|NumType=Card', POS: NUM},
    'NOUN__Definite=Ind|Gender=Masc': {'morph': 'Definite=Ind|Gender=Masc', POS: NOUN},
    'NOUN__Definite=Def|Number=Plur': {'morph': 'Definite=Def|Number=Plur', POS: NOUN},
    'ADJ__Number=Sing|VerbForm=Part': {'morph': 'Number=Sing|VerbForm=Part', POS: ADJ},
    'ADJ__Definite=Ind|Gender=Masc|Number=Sing|VerbForm=Part': {'morph': 'Definite=Ind|Gender=Masc|Number=Sing|VerbForm=Part', POS: ADJ},
    'NOUN__Abbr=Yes|Gender=Masc': {'morph': 'Abbr=Yes|Gender=Masc', POS: NOUN},
    'NOUN__Abbr=Yes|Case=Gen': {'morph': 'Abbr=Yes|Case=Gen', POS: NOUN},
    'VERB__Abbr=Yes|Mood=Ind|Tense=Pres|VerbForm=Fin': {'morph': 'Abbr=Yes|Mood=Ind|Tense=Pres|VerbForm=Fin', POS: VERB},
    'ADJ__Abbr=Yes|Degree=Pos': {'morph': 'Abbr=Yes|Degree=Pos', POS: ADJ},
    'NOUN__Case=Gen|Gender=Fem': {'morph': 'Case=Gen|Gender=Fem', POS: NOUN},
    'ADJ__Case=Gen|Degree=Cmp': {'morph': 'Case=Gen|Degree=Cmp', POS: ADJ},
    'ADJ__Definite=Ind|Degree=Pos|Gender=Masc|Number=Sing': {'morph': 'Definite=Ind|Degree=Pos|Gender=Masc|Number=Sing', POS: ADJ},
    'NOUN__Gender=Masc|Number=Sing': {'morph': 'Gender=Masc|Number=Sing', POS: NOUN},
}



# cython: infer_types
# coding: utf8
from __future__ import unicode_literals

from libc.string cimport memset

from .attrs cimport POS, IS_SPACE
from .attrs import LEMMA, intify_attrs
from .parts_of_speech cimport SPACE
from .parts_of_speech import IDS as POS_IDS
from .lexeme cimport Lexeme


def _normalize_props(props):
    """Transform deprecated string keys to correct names."""
    out = {}
    for key, value in props.items():
        if key == POS:
            if hasattr(value, 'upper'):
                value = value.upper()
            if value in POS_IDS:
                value = POS_IDS[value]
            out[key] = value
        elif isinstance(key, int):
            out[key] = value
        elif key.lower() == 'pos':
            out[POS] = POS_IDS[value.upper()]
        else:
            out[key] = value
    return out


cdef class Morphology:
    def __init__(self, StringStore string_store, tag_map, lemmatizer, exc=None):
        self.mem = Pool()
        self.strings = string_store
        # Add special space symbol. We prefix with underscore, to make sure it
        # always sorts to the end.
        space_attrs = tag_map.get('SP', {POS: SPACE})
        if '_SP' not in tag_map:
            self.strings.add('_SP')
            tag_map = dict(tag_map)
            tag_map['_SP'] = space_attrs
        self.tag_names = tuple(sorted(tag_map.keys()))
        self.tag_map = {}
        self.lemmatizer = lemmatizer
        self.n_tags = len(tag_map)
        self.reverse_index = {}

        self.rich_tags = <RichTagC*>self.mem.alloc(self.n_tags+1, sizeof(RichTagC))
        for i, (tag_str, attrs) in enumerate(sorted(tag_map.items())):
            self.strings.add(tag_str)
            self.tag_map[tag_str] = dict(attrs)
            attrs = _normalize_props(attrs)
            attrs = intify_attrs(attrs, self.strings, _do_deprecated=True)
            self.rich_tags[i].id = i
            self.rich_tags[i].name = self.strings.add(tag_str)
            self.rich_tags[i].morph = 0
            self.rich_tags[i].pos = attrs[POS]
            self.reverse_index[self.rich_tags[i].name] = i
        # Add a 'null' tag, which we can reference when assign morphology to
        # untagged tokens.
        self.rich_tags[self.n_tags].id = self.n_tags

        self._cache = PreshMapArray(self.n_tags)
        self.exc = {}
        if exc is not None:
            for (tag_str, orth_str), attrs in exc.items():
                self.add_special_case(tag_str, orth_str, attrs)

    def __reduce__(self):
        return (Morphology, (self.strings, self.tag_map, self.lemmatizer,
                             self.exc), None, None)

    cdef int assign_untagged(self, TokenC* token) except -1:
        """Set morphological attributes on a token without a POS tag. Uses
        the lemmatizer's lookup() method, which looks up the string in the
        table provided by the language data as lemma_lookup (if available).
        """
        if token.lemma == 0:
            orth_str = self.strings[token.lex.orth]
            lemma = self.lemmatizer.lookup(orth_str)
            token.lemma = self.strings.add(lemma)

    cdef int assign_tag(self, TokenC* token, tag) except -1:
        if isinstance(tag, basestring):
            tag = self.strings.add(tag)
        if tag in self.reverse_index:
            tag_id = self.reverse_index[tag]
            self.assign_tag_id(token, tag_id)
        else:
            token.tag = tag

    cdef int assign_tag_id(self, TokenC* token, int tag_id) except -1:
        if tag_id > self.n_tags:
            raise ValueError("Unknown tag ID: %s" % tag_id)
        # TODO: It's pretty arbitrary to put this logic here. I guess the
        # justification is that this is where the specific word and the tag
        # interact. Still, we should have a better way to enforce this rule, or
        # figure out why the statistical model fails. Related to Issue #220
        if Lexeme.c_check_flag(token.lex, IS_SPACE):
            tag_id = self.reverse_index[self.strings.add('_SP')]
        rich_tag = self.rich_tags[tag_id]
        analysis = <MorphAnalysisC*>self._cache.get(tag_id, token.lex.orth)
        if analysis is NULL:
            analysis = <MorphAnalysisC*>self.mem.alloc(1, sizeof(MorphAnalysisC))
            tag_str = self.strings[self.rich_tags[tag_id].name]
            analysis.tag = rich_tag
            analysis.lemma = self.lemmatize(analysis.tag.pos, token.lex.orth,
                                            self.tag_map.get(tag_str, {}))
            self._cache.set(tag_id, token.lex.orth, analysis)
        token.lemma = analysis.lemma
        token.pos = analysis.tag.pos
        token.tag = analysis.tag.name
        token.morph = analysis.tag.morph

    cdef int assign_feature(self, uint64_t* flags, univ_morph_t flag_id, bint value) except -1:
        cdef flags_t one = 1
        if value:
            flags[0] |= one << flag_id
        else:
            flags[0] &= ~(one << flag_id)

    def add_special_case(self, unicode tag_str, unicode orth_str, attrs,
                         force=False):
        """Add a special-case rule to the morphological analyser. Tokens whose
        tag and orth match the rule will receive the specified properties.

        tag (unicode): The part-of-speech tag to key the exception.
        orth (unicode): The word-form to key the exception.
        """
        # TODO: Currently we've assumed that we know the number of tags -- 
        # RichTagC is an array, and _cache is a PreshMapArray
        # This is really bad: it makes the morphology typed to the tagger
        # classes, which is all wrong.
        self.exc[(tag_str, orth_str)] = dict(attrs)
        tag = self.strings.add(tag_str)
        if tag not in self.reverse_index:
            return
        tag_id = self.reverse_index[tag]
        orth = self.strings[orth_str]
        cdef RichTagC rich_tag = self.rich_tags[tag_id]
        attrs = intify_attrs(attrs, self.strings, _do_deprecated=True)
        cached = <MorphAnalysisC*>self._cache.get(tag_id, orth)
        if cached is NULL:
            cached = <MorphAnalysisC*>self.mem.alloc(1, sizeof(MorphAnalysisC))
        elif force:
            memset(cached, 0, sizeof(cached[0]))
        else:
            raise ValueError(
                "Conflicting morphology exception for (%s, %s). Use "
                "force=True to overwrite." % (tag_str, orth_str))

        cached.tag = rich_tag
        # TODO: Refactor this to take arbitrary attributes.
        for name_id, value_id in attrs.items():
            if name_id == LEMMA:
                cached.lemma = value_id
            else:
                self.assign_feature(&cached.tag.morph, name_id, value_id)
        if cached.lemma == 0:
            cached.lemma = self.lemmatize(rich_tag.pos, orth, attrs)
        self._cache.set(tag_id, orth, <void*>cached)

    def load_morph_exceptions(self, dict exc):
        # Map (form, pos) to (lemma, rich tag)
        for tag_str, entries in exc.items():
            for form_str, attrs in entries.items():
                self.add_special_case(tag_str, form_str, attrs)

    def lemmatize(self, const univ_pos_t univ_pos, attr_t orth, morphology):
        if orth not in self.strings:
            return orth
        cdef unicode py_string = self.strings[orth]
        if self.lemmatizer is None:
            return self.strings.add(py_string.lower())
        cdef list lemma_strings
        cdef unicode lemma_string
        lemma_strings = self.lemmatizer(py_string, univ_pos, morphology)
        lemma_string = sorted(lemma_strings)[0]
        lemma = self.strings.add(lemma_string)
        return lemma


IDS = {
    "Animacy_anim": Animacy_anim,
    "Animacy_inam": Animacy_inam,
    "Aspect_freq": Aspect_freq,
    "Aspect_imp": Aspect_imp,
    "Aspect_mod": Aspect_mod,
    "Aspect_none": Aspect_none,
    "Aspect_perf": Aspect_perf,
    "Case_abe": Case_abe,
    "Case_abl": Case_abl,
    "Case_abs": Case_abs,
    "Case_acc": Case_acc,
    "Case_ade": Case_ade,
    "Case_all": Case_all,
    "Case_cau": Case_cau,
    "Case_com": Case_com,
    "Case_dat": Case_dat,
    "Case_del": Case_del,
    "Case_dis": Case_dis,
    "Case_ela": Case_ela,
    "Case_ess": Case_ess,
    "Case_gen": Case_gen,
    "Case_ill": Case_ill,
    "Case_ine": Case_ine,
    "Case_ins": Case_ins,
    "Case_loc": Case_loc,
    "Case_lat": Case_lat,
    "Case_nom": Case_nom,
    "Case_par": Case_par,
    "Case_sub": Case_sub,
    "Case_sup": Case_sup,
    "Case_tem": Case_tem,
    "Case_ter": Case_ter,
    "Case_tra": Case_tra,
    "Case_voc": Case_voc,
    "Definite_two": Definite_two,
    "Definite_def": Definite_def,
    "Definite_red": Definite_red,
    "Definite_cons": Definite_cons,  # U20
    "Definite_ind": Definite_ind,
    "Degree_cmp": Degree_cmp,
    "Degree_comp": Degree_comp,
    "Degree_none": Degree_none,
    "Degree_pos": Degree_pos,
    "Degree_sup": Degree_sup,
    "Degree_abs": Degree_abs,
    "Degree_com": Degree_com,
    "Degree_dim ": Degree_dim,  # du
    "Gender_com": Gender_com,
    "Gender_fem": Gender_fem,
    "Gender_masc": Gender_masc,
    "Gender_neut": Gender_neut,
    "Mood_cnd": Mood_cnd,
    "Mood_imp": Mood_imp,
    "Mood_ind": Mood_ind,
    "Mood_n": Mood_n,
    "Mood_pot": Mood_pot,
    "Mood_sub": Mood_sub,
    "Mood_opt": Mood_opt,
    "Negative_neg": Negative_neg,
    "Negative_pos": Negative_pos,
    "Negative_yes": Negative_yes,
    "Polarity_neg": Polarity_neg,  # U20
    "Polarity_pos": Polarity_pos,  # U20
    "Number_com": Number_com,
    "Number_dual": Number_dual,
    "Number_none": Number_none,
    "Number_plur": Number_plur,
    "Number_sing": Number_sing,
    "Number_ptan ": Number_ptan,  # bg
    "Number_count ": Number_count,  # bg
    "NumType_card": NumType_card,
    "NumType_dist": NumType_dist,
    "NumType_frac": NumType_frac,
    "NumType_gen": NumType_gen,
    "NumType_mult": NumType_mult,
    "NumType_none": NumType_none,
    "NumType_ord": NumType_ord,
    "NumType_sets": NumType_sets,
    "Person_one": Person_one,
    "Person_two": Person_two,
    "Person_three": Person_three,
    "Person_none": Person_none,
    "Poss_yes": Poss_yes,
    "PronType_advPart": PronType_advPart,
    "PronType_art": PronType_art,
    "PronType_default": PronType_default,
    "PronType_dem": PronType_dem,
    "PronType_ind": PronType_ind,
    "PronType_int": PronType_int,
    "PronType_neg": PronType_neg,
    "PronType_prs": PronType_prs,
    "PronType_rcp": PronType_rcp,
    "PronType_rel": PronType_rel,
    "PronType_tot": PronType_tot,
    "PronType_clit": PronType_clit,
    "PronType_exc ": PronType_exc,  # es, ca, it, fa,
    "Reflex_yes": Reflex_yes,
    "Tense_fut": Tense_fut,
    "Tense_imp": Tense_imp,
    "Tense_past": Tense_past,
    "Tense_pres": Tense_pres,
    "VerbForm_fin": VerbForm_fin,
    "VerbForm_ger": VerbForm_ger,
    "VerbForm_inf": VerbForm_inf,
    "VerbForm_none": VerbForm_none,
    "VerbForm_part": VerbForm_part,
    "VerbForm_partFut": VerbForm_partFut,
    "VerbForm_partPast": VerbForm_partPast,
    "VerbForm_partPres": VerbForm_partPres,
    "VerbForm_sup": VerbForm_sup,
    "VerbForm_trans": VerbForm_trans,
    "VerbForm_conv": VerbForm_conv,  # U20
    "VerbForm_gdv ": VerbForm_gdv,  # la,
    "Voice_act": Voice_act,
    "Voice_cau": Voice_cau,
    "Voice_pass": Voice_pass,
    "Voice_mid ": Voice_mid,  # gkc,
    "Voice_int ": Voice_int,  # hb,
    "Abbr_yes ": Abbr_yes,  # cz, fi, sl, U,
    "AdpType_prep ": AdpType_prep,  # cz, U,
    "AdpType_post ": AdpType_post,  # U,
    "AdpType_voc ": AdpType_voc,  # cz,
    "AdpType_comprep ": AdpType_comprep,  # cz,
    "AdpType_circ ": AdpType_circ,  # U,
    "AdvType_man": AdvType_man,
    "AdvType_loc": AdvType_loc,
    "AdvType_tim": AdvType_tim,
    "AdvType_deg": AdvType_deg,
    "AdvType_cau": AdvType_cau,
    "AdvType_mod": AdvType_mod,
    "AdvType_sta": AdvType_sta,
    "AdvType_ex": AdvType_ex,
    "AdvType_adadj": AdvType_adadj,
    "ConjType_oper ": ConjType_oper,  # cz, U,
    "ConjType_comp ": ConjType_comp,  # cz, U,
    "Connegative_yes ": Connegative_yes,  # fi,
    "Derivation_minen ": Derivation_minen,  # fi,
    "Derivation_sti ": Derivation_sti,  # fi,
    "Derivation_inen ": Derivation_inen,  # fi,
    "Derivation_lainen ": Derivation_lainen,  # fi,
    "Derivation_ja ": Derivation_ja,  # fi,
    "Derivation_ton ": Derivation_ton,  # fi,
    "Derivation_vs ": Derivation_vs,  # fi,
    "Derivation_ttain ": Derivation_ttain,  # fi,
    "Derivation_ttaa ": Derivation_ttaa,  # fi,
    "Echo_rdp ": Echo_rdp,  # U,
    "Echo_ech ": Echo_ech,  # U,
    "Foreign_foreign ": Foreign_foreign,  # cz, fi, U,
    "Foreign_fscript ": Foreign_fscript,  # cz, fi, U,
    "Foreign_tscript ": Foreign_tscript,  # cz, U,
    "Foreign_yes ": Foreign_yes,  # sl,
    "Gender_dat_masc ": Gender_dat_masc,  # bq, U,
    "Gender_dat_fem ": Gender_dat_fem,  # bq, U,
    "Gender_erg_masc ": Gender_erg_masc,  # bq,
    "Gender_erg_fem ": Gender_erg_fem,  # bq,
    "Gender_psor_masc ": Gender_psor_masc,  # cz, sl, U,
    "Gender_psor_fem ": Gender_psor_fem,  # cz, sl, U,
    "Gender_psor_neut ": Gender_psor_neut,  # sl,
    "Hyph_yes ": Hyph_yes,  # cz, U,
    "InfForm_one ": InfForm_one,  # fi,
    "InfForm_two ": InfForm_two,  # fi,
    "InfForm_three ": InfForm_three,  # fi,
    "NameType_geo ": NameType_geo,  # U, cz,
    "NameType_prs ": NameType_prs,  # U, cz,
    "NameType_giv ": NameType_giv,  # U, cz,
    "NameType_sur ": NameType_sur,  # U, cz,
    "NameType_nat ": NameType_nat,  # U, cz,
    "NameType_com ": NameType_com,  # U, cz,
    "NameType_pro ": NameType_pro,  # U, cz,
    "NameType_oth ": NameType_oth,  # U, cz,
    "NounType_com ": NounType_com,  # U,
    "NounType_prop ": NounType_prop,  # U,
    "NounType_class ": NounType_class,  # U,
    "Number_abs_sing ": Number_abs_sing,  # bq, U,
    "Number_abs_plur ": Number_abs_plur,  # bq, U,
    "Number_dat_sing ": Number_dat_sing,  # bq, U,
    "Number_dat_plur ": Number_dat_plur,  # bq, U,
    "Number_erg_sing ": Number_erg_sing,  # bq, U,
    "Number_erg_plur ": Number_erg_plur,  # bq, U,
    "Number_psee_sing ": Number_psee_sing,  # U,
    "Number_psee_plur ": Number_psee_plur,  # U,
    "Number_psor_sing ": Number_psor_sing,  # cz, fi, sl, U,
    "Number_psor_plur ": Number_psor_plur,  # cz, fi, sl, U,
    "NumForm_digit ": NumForm_digit,  # cz, sl, U,
    "NumForm_roman ": NumForm_roman,  # cz, sl, U,
    "NumForm_word ": NumForm_word,  # cz, sl, U,
    "NumValue_one ": NumValue_one,  # cz, U,
    "NumValue_two ": NumValue_two,  # cz, U,
    "NumValue_three ": NumValue_three,  # cz, U,
    "PartForm_pres ": PartForm_pres,  # fi,
    "PartForm_past ": PartForm_past,  # fi,
    "PartForm_agt ": PartForm_agt,  # fi,
    "PartForm_neg ": PartForm_neg,  # fi,
    "PartType_mod ": PartType_mod,  # U,
    "PartType_emp ": PartType_emp,  # U,
    "PartType_res ": PartType_res,  # U,
    "PartType_inf ": PartType_inf,  # U,
    "PartType_vbp ": PartType_vbp,  # U,
    "Person_abs_one ": Person_abs_one,  # bq, U,
    "Person_abs_two ": Person_abs_two,  # bq, U,
    "Person_abs_three ": Person_abs_three,  # bq, U,
    "Person_dat_one ": Person_dat_one,  # bq, U,
    "Person_dat_two ": Person_dat_two,  # bq, U,
    "Person_dat_three ": Person_dat_three,  # bq, U,
    "Person_erg_one ": Person_erg_one,  # bq, U,
    "Person_erg_two ": Person_erg_two,  # bq, U,
    "Person_erg_three ": Person_erg_three,  # bq, U,
    "Person_psor_one ": Person_psor_one,  # fi, U,
    "Person_psor_two ": Person_psor_two,  # fi, U,
    "Person_psor_three ": Person_psor_three,  # fi, U,
    "Polite_inf ": Polite_inf,  # bq, U,
    "Polite_pol ": Polite_pol,  # bq, U,
    "Polite_abs_inf ": Polite_abs_inf,  # bq, U,
    "Polite_abs_pol ": Polite_abs_pol,  # bq, U,
    "Polite_erg_inf ": Polite_erg_inf,  # bq, U,
    "Polite_erg_pol ": Polite_erg_pol,  # bq, U,
    "Polite_dat_inf ": Polite_dat_inf,  # bq, U,
    "Polite_dat_pol ": Polite_dat_pol,  # bq, U,
    "Prefix_yes ": Prefix_yes,  # U,
    "PrepCase_npr ": PrepCase_npr,  # cz,
    "PrepCase_pre ": PrepCase_pre,  # U,
    "PunctSide_ini ": PunctSide_ini,  # U,
    "PunctSide_fin ": PunctSide_fin,  # U,
    "PunctType_peri ": PunctType_peri,  # U,
    "PunctType_qest ": PunctType_qest,  # U,
    "PunctType_excl ": PunctType_excl,  # U,
    "PunctType_quot ": PunctType_quot,  # U,
    "PunctType_brck ": PunctType_brck,  # U,
    "PunctType_comm ": PunctType_comm,  # U,
    "PunctType_colo ": PunctType_colo,  # U,
    "PunctType_semi ": PunctType_semi,  # U,
    "PunctType_dash ": PunctType_dash,  # U,
    "Style_arch ": Style_arch,  # cz, fi, U,
    "Style_rare ": Style_rare,  # cz, fi, U,
    "Style_poet ": Style_poet,  # cz, U,
    "Style_norm ": Style_norm,  # cz, U,
    "Style_coll ": Style_coll,  # cz, U,
    "Style_vrnc ": Style_vrnc,  # cz, U,
    "Style_sing ": Style_sing,  # cz, U,
    "Style_expr ": Style_expr,  # cz, U,
    "Style_derg ": Style_derg,  # cz, U,
    "Style_vulg ": Style_vulg,  # cz, U,
    "Style_yes ": Style_yes,  # fi, U,
    "StyleVariant_styleShort ": StyleVariant_styleShort,  # cz,
    "StyleVariant_styleBound ": StyleVariant_styleBound,  # cz, sl,
    "VerbType_aux ": VerbType_aux,  # U,
    "VerbType_cop ": VerbType_cop,  # U,
    "VerbType_mod ": VerbType_mod,  # U,
    "VerbType_light ": VerbType_light,  # U,
}


NAMES = [key for key, value in sorted(IDS.items(), key=lambda item: item[1])]
# Unfortunate hack here, to work around problem with long cpdef enum
# (which is generating an enormous amount of C++ in Cython 0.24+)
# We keep the enum cdef, and just make sure the names are available to Python
locals().update(IDS)



# coding: utf-8
from __future__ import unicode_literals

import pytest


@pytest.mark.models('fr')
def test_lemmatizer_verb(FR):
    tokens = FR("Qu'est-ce que tu fais?")
    assert tokens[0].lemma_ == "que"
    assert tokens[1].lemma_ == "être"
    assert tokens[5].lemma_ == "faire"


@pytest.mark.models('fr')
@pytest.mark.xfail(reason="sont tagged as AUX")
def test_lemmatizer_noun_verb_2(FR):
    tokens = FR("Les abaissements de température sont gênants.")
    assert tokens[4].lemma_ == "être"


@pytest.mark.models('fr')
@pytest.mark.xfail(reason="Costaricienne TAG is PROPN instead of NOUN and spacy don't lemmatize PROPN")
def test_lemmatizer_noun(FR):
    tokens = FR("il y a des Costaricienne.")
    assert tokens[4].lemma_ == "Costaricain"


@pytest.mark.models('fr')
def test_lemmatizer_noun_2(FR):
    tokens = FR("Les abaissements de température sont gênants.")
    assert tokens[1].lemma_ == "abaissement"
    assert tokens[5].lemma_ == "gênant"



# encoding: utf8
"""
Tokenizer Exceptions.
Source: https://forkortelse.dk/ and various others.
"""

from __future__ import unicode_literals

from ...symbols import ORTH, LEMMA, NORM, TAG, PUNCT


_exc = {}

# Abbreviations for weekdays "søn." (for "søndag") as well as "Tor." and "Tors."
# (for "torsdag") are left out because they are ambiguous. The same is the case
# for abbreviations "jul." and "Jul." ("juli").
for exc_data in [
        {ORTH: "Kbh.", LEMMA: "København", NORM: "København"},
        {ORTH: "jan.", LEMMA: "januar"},
        {ORTH: "febr.", LEMMA: "februar"},
        {ORTH: "feb.", LEMMA: "februar"},
        {ORTH: "mar.", LEMMA: "marts"},
        {ORTH: "apr.", LEMMA: "april"},
        {ORTH: "jun.", LEMMA: "juni"},
        {ORTH: "aug.", LEMMA: "august"},
        {ORTH: "sept.", LEMMA: "september"},
        {ORTH: "sep.", LEMMA: "september"},
        {ORTH: "okt.", LEMMA: "oktober"},
        {ORTH: "nov.", LEMMA: "november"},
        {ORTH: "dec.", LEMMA: "december"},
        {ORTH: "man.", LEMMA: "mandag"},
        {ORTH: "tirs.", LEMMA: "tirsdag"},
        {ORTH: "ons.", LEMMA: "onsdag"},
        {ORTH: "tor.", LEMMA: "torsdag"},
        {ORTH: "tors.", LEMMA: "torsdag"},
        {ORTH: "fre.", LEMMA: "fredag"},
        {ORTH: "lør.", LEMMA: "lørdag"},
        {ORTH: "Jan.", LEMMA: "januar"},
        {ORTH: "Febr.", LEMMA: "februar"},
        {ORTH: "Feb.", LEMMA: "februar"},
        {ORTH: "Mar.", LEMMA: "marts"},
        {ORTH: "Apr.", LEMMA: "april"},
        {ORTH: "Jun.", LEMMA: "juni"},
        {ORTH: "Aug.", LEMMA: "august"},
        {ORTH: "Sept.", LEMMA: "september"},
        {ORTH: "Sep.", LEMMA: "september"},
        {ORTH: "Okt.", LEMMA: "oktober"},
        {ORTH: "Nov.", LEMMA: "november"},
        {ORTH: "Dec.", LEMMA: "december"},
        {ORTH: "Man.", LEMMA: "mandag"},
        {ORTH: "Tirs.", LEMMA: "tirsdag"},
        {ORTH: "Ons.", LEMMA: "onsdag"},
        {ORTH: "Fre.", LEMMA: "fredag"},
        {ORTH: "Lør.", LEMMA: "lørdag"}]:
    _exc[exc_data[ORTH]] = [exc_data]


# Specified case only
for orth in [
        "diam.", "ib.", "mia.", "mik.", "pers.", "A.D.", "A/S", "B.C.", "BK.",
        "Dr.", "Boul.", "Chr.", "Dronn.", "H.K.H.", "H.M.", "Hf.", "i/s", "I/S",
        "Kprs.", "L.A.", "Ll.", "m/s", "M/S", "Mag.", "Mr.", "Ndr.", "Ph.d.",
        "Prs.", "Rcp.", "Sdr.", "Skt.", "Spl.", "Vg."]:
    _exc[orth] = [{ORTH: orth}]


for orth in [
        "aarh.", "ac.", "adj.", "adr.", "adsk.", "adv.", "afb.", "afd.", "afg.",
        "afk.", "afs.", "aht.", "alg.", "alk.", "alm.", "amer.", "ang.", "ank.",
        "anl.", "anv.", "arb.", "arr.", "att.", "bd.", "bdt.", "beg.", "begr.",
        "beh.", "bet.", "bev.", "bhk.", "bib.", "bibl.", "bidr.", "bildl.",
        "bill.", "biol.", "bk.", "bl.", "bl.a.", "borgm.", "br.", "brolægn.",
        "bto.", "bygn.", "ca.", "cand.", "d.d.", "d.m.", "d.s.", "d.s.s.",
        "d.y.", "d.å.", "d.æ.", "dagl.", "dat.", "dav.", "def.", "dek.", "dep.",
        "desl.", "dir.", "disp.", "distr.", "div.", "dkr.", "dl.", "do.",
        "dobb.", "dr.h.c", "dr.phil.", "ds.", "dvs.", "e.b.", "e.l.", "e.o.",
        "e.v.t.", "eftf.", "eftm.", "egl.", "eks.", "eksam.", "ekskl.", "eksp.",
        "ekspl.", "el.lign.", "emer.", "endv.", "eng.", "enk.", "etc.", "etym.",
        "eur.", "evt.", "exam.", "f.eks.", "f.m.", "f.n.", "f.o.", "f.o.m.",
        "f.s.v.", "f.t.", "f.v.t.", "f.å.", "fa.", "fakt.", "fam.", "ff.",
        "fg.", "fhv.", "fig.", "filol.", "filos.", "fl.", "flg.", "fm.", "fmd.",
        "fol.", "forb.", "foreg.", "foren.", "forf.", "fork.", "forr.", "fors.",
        "forsk.", "forts.", "fr.", "fr.u.", "frk.", "fsva.", "fuldm.", "fung.",
        "fx.", "fys.", "fær.", "g.d.", "g.m.", "gd.", "gdr.", "genuds.", "gl.",
        "gn.", "gns.", "gr.", "grdl.", "gross.", "h.a.", "h.c.", "hdl.",
        "henv.", "hhv.", "hj.hj.", "hj.spl.", "hort.", "hosp.", "hpl.", "hr.",
        "hrs.", "hum.", "hvp.", "i.e.", "id.", "if.", "iflg.", "ifm.", "ift.",
        "iht.", "ill.", "indb.", "indreg.", "inf.", "ing.", "inh.", "inj.",
        "inkl.", "insp.", "instr.", "isl.", "istf.", "it.", "ital.", "iv.",
        "jap.", "jf.", "jfr.", "jnr.", "j.nr.", "jr.", "jur.", "jvf.", "kap.",
        "kbh.", "kem.", "kgl.", "kl.", "kld.", "knsp.", "komm.", "kons.",
        "korr.", "kp.", "kr.", "kst.", "kt.", "ktr.", "kv.", "kvt.", "l.c.",
        "lab.", "lat.", "lb.m.", "lb.nr.", "lejl.", "lgd.", "lic.", "lign.",
        "lin.", "ling.merc.", "litt.", "loc.cit.", "lok.", "lrs.", "ltr.",
        "m.a.o.", "m.fl.", "m.m.", "m.v.", "m.v.h.", "maks.", "md.", "mdr.",
        "mdtl.", "mezz.", "mfl.", "m.h.p.", "m.h.t.", "mht.", "mill.", "mio.",
        "modt.", "mrk.", "mul.", "mv.", "n.br.", "n.f.", "nb.", "nedenst.",
        "nl.", "nr.", "nto.", "nuv.", "o/m", "o.a.", "o.fl.", "o.h.", "o.l.",
        "o.lign.", "o.m.a.", "o.s.fr.", "obl.", "obs.", "odont.", "oecon.",
        "off.", "ofl.", "omg.", "omkr.", "omr.", "omtr.", "opg.", "opl.",
        "opr.", "org.", "orig.", "osv.", "ovenst.", "overs.", "ovf.", "p.a.",
        "p.b.a", "p.b.v", "p.c.", "p.m.", "p.m.v.", "p.n.", "p.p.", "p.p.s.",
        "p.s.", "p.t.", "p.v.a.", "p.v.c.", "pag.", "pass.", "pcs.", "pct.",
        "pd.", "pens.", "pft.", "pg.", "pga.", "pgl.", "pinx.", "pk.", "pkt.",
        "polit.", "polyt.", "pos.", "pp.", "ppm.", "pr.", "prc.", "priv.",
        "prod.", "prof.", "pron.", "præd.", "præf.", "præt.", "psych.", "pt.",
        "pæd.", "q.e.d.", "rad.", "red.", "ref.", "reg.", "regn.", "rel.",
        "rep.", "repr.", "resp.", "rest.", "rm.", "rtg.", "russ.", "s.br.",
        "s.d.", "s.f.", "s.m.b.a.", "s.u.", "s.å.", "sa.", "sb.", "sc.",
        "scient.", "scil.", "sek.", "sekr.", "self.", "sem.", "shj.", "sign.",
        "sing.", "sj.", "skr.", "slutn.", "sml.", "smp.", "snr.", "soc.",
        "soc.dem.", "sp.", "spec.", "spm.", "spr.", "spsk.", "statsaut.", "st.",
        "stk.", "str.", "stud.", "subj.", "subst.", "suff.", "sup.", "suppl.",
        "sv.", "såk.", "sædv.", "t/r", "t.h.", "t.o.", "t.o.m.", "t.v.", "tbl.",
        "tcp/ip", "td.", "tdl.", "tdr.", "techn.", "tekn.", "temp.", "th.",
        "theol.", "tidl.", "tilf.", "tilh.", "till.", "tilsv.", "tjg.", "tkr.",
        "tlf.", "tlgr.", "tr.", "trp.", "tsk.", "tv.", "ty.", "u/b", "udb.",
        "udbet.", "ugtl.", "undt.", "v.f.", "vb.", "vedk.", "vedl.", "vedr.",
        "vejl.", "vh.", "vha.", "vs.", "vsa.", "vær.", "zool.", "ø.lgd.",
        "øvr.", "årg.", "årh."]:
    _exc[orth] = [{ORTH: orth}]
    capitalized = orth.capitalize()
    _exc[capitalized] = [{ORTH: capitalized}]

for exc_data in [
        {ORTH: "s'gu", LEMMA: "s'gu", NORM: "s'gu"},
        {ORTH: "S'gu", LEMMA: "s'gu", NORM: "s'gu"},
        {ORTH: "sgu'", LEMMA: "s'gu", NORM: "s'gu"},
        {ORTH: "Sgu'", LEMMA: "s'gu", NORM: "s'gu"},
        {ORTH: "sku'", LEMMA: "skal", NORM: "skulle"},
        {ORTH: "ku'", LEMMA: "kan", NORM: "kunne"},
        {ORTH: "Ku'", LEMMA: "kan", NORM: "kunne"},
        {ORTH: "ka'", LEMMA: "kan", NORM: "kan"},
        {ORTH: "Ka'", LEMMA: "kan", NORM: "kan"},
        {ORTH: "gi'", LEMMA: "give", NORM: "giv"},
        {ORTH: "Gi'", LEMMA: "give", NORM: "giv"},
        {ORTH: "li'", LEMMA: "lide", NORM: "lide"},
        {ORTH: "ha'", LEMMA: "have", NORM: "have"},
        {ORTH: "Ha'", LEMMA: "have", NORM: "have"},
        {ORTH: "ik'", LEMMA: "ikke", NORM: "ikke"},
        {ORTH: "Ik'", LEMMA: "ikke", NORM: "ikke"}]:
    _exc[exc_data[ORTH]] = [exc_data]


# Dates
for h in range(1, 31 + 1):
    for period in ["."]:
        _exc["%d%s" % (h, period)] = [{ORTH: "%d." % h}]

_custom_base_exc = {
    "i.": [
        {ORTH: "i", LEMMA: "i", NORM: "i"},
        {ORTH: ".", TAG: PUNCT}]
}
_exc.update(_custom_base_exc)

TOKENIZER_EXCEPTIONS = _exc



# coding: utf8
from __future__ import unicode_literals

_exc = {
    "Rp": "$",
    "IDR": "$",
    "RMB": "$",
    "USD": "$",
    "AUD": "$",
    "GBP": "$",
}

NORM_EXCEPTIONS = {}

for string, norm in _exc.items():
    NORM_EXCEPTIONS[string] = norm
    NORM_EXCEPTIONS[string.title()] = norm



# coding: utf-8
from __future__ import unicode_literals

import pytest


@pytest.mark.parametrize('text', ["would've"])
def test_issue1758(en_tokenizer, text):
    """Test that "would've" is handled by the English tokenizer exceptions."""
    tokens = en_tokenizer(text)
    assert len(tokens) == 2
    assert tokens[0].tag_ == "MD"
    assert tokens[1].lemma_ == "have"
